// PeakSearch.cpp
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include <math.h>
#include <float.h>
#include <stddef.h>
#include "PeakSearch.h"
#include "..\IniFile.h"
#include "..\WinSub.h"

extern CIniFile *IniFile;

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

extern int lstsq(double *x, double *y, int n, int m, double *c);
extern int ToClip(char *Title, int ndata, int ny, double *x, double *y[], double offset);

CPeakSearch::CPeakSearch()
{
	m_iPrt          = IniFile->GetInt("SetupLSQ", "iPrt", 1);
	m_iPrtIntvl     = IniFile->WriteInt("SetupLSQ", "iPrtIntvl", 10);
	m_BGOrder       = 4;
//	m_SpectrumKind  = PEAKSEARCH_SPECTRUMKIND_INVALID;
	m_SpectrumKind  = PEAKSEARCH_SPECTRUMKIND_XRD;
	m_pOriginalData = new CVariousData;
	char str[256];
	IniFile->GetString("SetupLSQ", "XRDLSQProgram", "", str);
	m_XRDLSQProgram = str;
	IniFile->GetString("SetupLSQ", "RamanLSQProgram", "", str);
	m_RamanLSQProgram = str;
	m_MaxIter = IniFile->GetInt("SetupLSQ", "MaxIteration", 30);
	m_ConvergenceCriteria = IniFile->GetDouble("SetupLSQ", "ConvergenceCriteria", 0.01);
	m_DumpingFactor = IniFile->GetDouble("SetupLSQ", "DumpingFactor", 0.3);

	m_QRange = IniFile->GetDouble("LSQ", "QRange", 5);
	m_IRatio = IniFile->GetDouble("LSQ", "IRatio", 0.5);
	m_LKa1   = IniFile->GetDouble("LSQ", "LKa1", 1.54056);
	m_LKa2   = IniFile->GetDouble("LSQ", "LKa2", 1.54439);

	m_Position      = NULL;
	m_CTotal        = NULL;
	m_GaussFraction = NULL;
	m_CGauss        = NULL;
	m_FWHMGauss     = NULL;
	m_CLorentz      = NULL;
	m_FWHMLorentz   = NULL;
	m_FWHMGLRatio   = NULL;
	m_Aniso         = NULL;
	m_OptimizePosition      = NULL;
	m_OptimizeCGauss        = NULL;
	m_OptimizeFWHMGauss     = NULL;
	m_OptimizeCLorentz      = NULL;
	m_OptimizeFWHMLorentz   = NULL;
	m_OptimizeFWHMGLRatio   = NULL;
	m_OptimizeAniso         = NULL;
	m_OptimizeCTotal        = NULL;
	m_OptimizeGaussFraction = NULL;
	m_DiffPosition      = NULL;
	m_DiffCTotal        = NULL;
	m_DiffGaussFraction = NULL;
	m_DiffCGauss        = NULL;
	m_DiffFWHMGauss     = NULL;
	m_DiffCLorentz      = NULL;
	m_DiffFWHMLorentz   = NULL;
	m_DiffFWHMGLRatio   = NULL;
	m_DiffAniso         = NULL;
	for(int i = 0 ; i < 5 ; i++)
	{
		m_OptimizeBGb[i] = 1;
		m_DiffBGb[i] = 0.0;
	}
	SetMaxPeaks(PEAKSEARCH_MAXPEAKS);
}

CPeakSearch::~CPeakSearch()
{
	if(m_pOriginalData) delete m_pOriginalData;
	if(m_Position)      delete []m_Position;
	if(m_CTotal)        delete []m_CTotal;
	if(m_GaussFraction) delete []m_GaussFraction;
	if(m_CGauss)        delete []m_CGauss;
	if(m_FWHMGauss)     delete []m_FWHMGauss;
	if(m_CLorentz)      delete []m_CLorentz;
	if(m_FWHMLorentz)   delete []m_FWHMLorentz;
	if(m_FWHMGLRatio)   delete []m_FWHMGLRatio;
	if(m_Aniso)         delete []m_Aniso;
	if(m_OptimizePosition)      delete []m_OptimizePosition;
	if(m_OptimizeCTotal)        delete []m_OptimizeCTotal;
	if(m_OptimizeGaussFraction) delete []m_OptimizeGaussFraction;
	if(m_OptimizeCGauss)        delete []m_OptimizeCGauss;
	if(m_OptimizeFWHMGauss)     delete []m_OptimizeFWHMGauss;
	if(m_OptimizeCLorentz)      delete []m_OptimizeCLorentz;
	if(m_OptimizeFWHMLorentz)   delete []m_OptimizeFWHMLorentz;
	if(m_OptimizeFWHMGLRatio)   delete []m_OptimizeFWHMGLRatio;
	if(m_OptimizeAniso)         delete []m_OptimizeAniso;
	if(m_DiffPosition)      delete []m_DiffPosition;
	if(m_DiffCTotal)        delete []m_DiffCTotal;
	if(m_DiffGaussFraction) delete []m_DiffGaussFraction;
	if(m_DiffCGauss)        delete []m_DiffCGauss;
	if(m_DiffFWHMGauss)     delete []m_DiffFWHMGauss;
	if(m_DiffCLorentz)      delete []m_DiffCLorentz;
	if(m_DiffFWHMLorentz)   delete []m_DiffFWHMLorentz;
	if(m_DiffFWHMGLRatio)   delete []m_DiffFWHMGLRatio;
}

int CPeakSearch::SetMaxPeaks(int MaxPeaks)
{
	m_MaxPeaks = MaxPeaks;
	m_nPeaks = 0;
	if(m_Position) delete []m_Position;
	m_Position = new double [m_MaxPeaks];
	if(m_CTotal) delete []m_CTotal;
	m_CTotal = new double [m_MaxPeaks];
	if(m_GaussFraction) delete []m_GaussFraction;
	m_GaussFraction = new double [m_MaxPeaks]; 
	if(m_CGauss) delete []m_CGauss;
	m_CGauss = new double [m_MaxPeaks];
	if(m_FWHMGauss) delete []m_FWHMGauss;
	m_FWHMGauss = new double [m_MaxPeaks];
	if(m_CLorentz) delete []m_CLorentz;
	m_CLorentz = new double [m_MaxPeaks];
	if(m_FWHMLorentz) delete []m_FWHMLorentz;
	m_FWHMLorentz = new double [m_MaxPeaks];
	if(m_FWHMGLRatio) delete []m_FWHMGLRatio;
	m_FWHMGLRatio = new double [m_MaxPeaks];
	if(m_Aniso) delete []m_Aniso;
	m_Aniso = new double [m_MaxPeaks];

	if(m_OptimizePosition) delete []m_OptimizePosition;
	m_OptimizePosition = new int [m_MaxPeaks];
	if(m_OptimizeCTotal) delete []m_OptimizeCTotal;
	m_OptimizeCTotal = new int [m_MaxPeaks];
	if(m_OptimizeGaussFraction) delete []m_OptimizeGaussFraction;
	m_OptimizeGaussFraction = new int [m_MaxPeaks];
	if(m_OptimizeCGauss) delete []m_OptimizeCGauss;
	m_OptimizeCGauss = new int [m_MaxPeaks];
	if(m_OptimizeFWHMGauss) delete []m_OptimizeFWHMGauss;
	m_OptimizeFWHMGauss = new int [m_MaxPeaks];
	if(m_OptimizeCLorentz) delete []m_OptimizeCLorentz;
	m_OptimizeCLorentz = new int [m_MaxPeaks];
	if(m_OptimizeFWHMLorentz) delete []m_OptimizeFWHMLorentz;
	m_OptimizeFWHMLorentz = new int [m_MaxPeaks];
	if(m_OptimizeFWHMGLRatio) delete []m_OptimizeFWHMGLRatio;
	m_OptimizeFWHMGLRatio = new int [m_MaxPeaks];
	if(m_OptimizeAniso) delete []m_OptimizeAniso;
	m_OptimizeAniso = new int [m_MaxPeaks];

	if(m_DiffPosition) delete []m_DiffPosition;
	m_DiffPosition = new double [m_MaxPeaks];
	if(m_DiffCTotal) delete []m_DiffCTotal;
	m_DiffCTotal = new double [m_MaxPeaks];
	if(m_DiffGaussFraction) delete []m_DiffGaussFraction;
	m_DiffGaussFraction = new double [m_MaxPeaks]; 
	if(m_DiffCGauss) delete []m_DiffCGauss;
	m_DiffCGauss = new double [m_MaxPeaks];
	if(m_DiffFWHMGauss) delete []m_DiffFWHMGauss;
	m_DiffFWHMGauss = new double [m_MaxPeaks];
	if(m_DiffCLorentz) delete []m_DiffCLorentz;
	m_DiffCLorentz = new double [m_MaxPeaks];
	if(m_DiffFWHMLorentz) delete []m_DiffFWHMLorentz;
	m_DiffFWHMLorentz = new double [m_MaxPeaks];
	if(m_DiffFWHMGLRatio) delete []m_DiffFWHMGLRatio;
	m_DiffFWHMGLRatio = new double [m_MaxPeaks];
	if(m_DiffAniso) delete []m_DiffAniso;
	m_DiffAniso = new double [m_MaxPeaks];
	return m_MaxPeaks;
}

int CPeakSearch::RemoveAll()
{
	for(int i = 0 ; i < 5 ; i++)
	{
		m_BGb[i] = 0.0;
		m_OptimizeBGb[i] = 1;
	}
	m_OptimizeBGb[2] = 0;
	m_OptimizeBGb[3] = 0;
	m_OptimizeBGb[4] = 0;
	return m_nPeaks = 0;
}

int CPeakSearch::CalcPeak(int iCurve)
{
	if(m_SpectrumKind == PEAKSEARCH_SPECTRUMKIND_XRD)
		return CalcPeakFitPeak(iCurve);
	else if(m_SpectrumKind == PEAKSEARCH_SPECTRUMKIND_RAMAN)
	{
		if(IsUsePeakFit(m_RamanLSQProgram))
			return CalcRmnFitPeak(iCurve, true);
		else
			return CalcRmnFitPeak(iCurve, false);
	}
	return -1;
}

int CPeakSearch::CalcRmnFitPeak(int iCurve, bool IsUsePeakFitBG)
{
	double f1;
	double Qmin, Qmax, qw, Q, xx1, Q1;
	int i0, i1, i, j;
	double Qa;
	double val, coeL;
	const double coeG = 2.772588722;
	double Lf, Gf, rf;

	CGraphData *pData = m_pOriginalData;
	if(pData == NULL) return -1;
	int datan = pData->nData();
	m_BG.SetDataNumber(datan);
	m_Total.SetDataNumber(datan);
	for(i = 0 ; i < datan ; i++)
	{
		m_BG.SetX(i, pData->X(i));
		m_BG.SetY(i, 0.0);
		m_Total.SetX(i, pData->X(i));
	}
	m_Total.m_LineLineColor = RGB(0xff, 0, 0);
	m_Total.SetFileName(pData->FileName());
	m_Total.SetComment("(Total)" + pData->FileName());
	m_BG.m_LineLineColor = RGB(0, 0xff, 0xff);
	m_BG.SetFileName(pData->FileName());
	m_BG.SetComment("(Background)" + pData->FileName());
	if(m_QRange < 0.1) m_QRange = 0.1;
	if(IsUsePeakFitBG)
	{
		for(i = 0 ; i < datan ; i++)
		{
			Q = m_BG.X(i);
			val = m_BGb[0] + (m_BGb[1] + (m_BGb[2] + (m_BGb[3] + m_BGb[4]*Q) * Q) * Q) * Q;
			m_BG.SetY(i, val);
			m_Total.SetY(i, val);
		}
	}
	else
	{
		double Cent = (m_OptRegionX0 + m_OptRegionX1)/2.0;
		double Rng  = (m_OptRegionX1 - m_OptRegionX0)/2.0;
		for(i = 0 ; i < datan ; i++)
		{
			Q = m_BG.X(i);
			val = m_BGb[0];
			for(int j = 1 ; j <= m_BGOrder ; j++)
			{
				val += m_BGb[j] * pow((Q-Cent)/Rng, j);
			}
			m_BG.SetY(i, val);
			m_Total.SetY(i, val);
		}
	}
	m_Curves.RemoveAllData();
	for(i = 0 ; i < m_nPeaks ; i++)
	{
		CGraphData *pCurve = m_Curves.AddGraph(datan);
		if(pCurve == NULL)
		{
			CString s;
			s.Format("Can't create curve #%d with datan=%d", i, datan);
			MessageBox(NULL, s, "Error", MB_OK);
			return -1;
		}
		for(j = 0 ; j < datan ; j++)
		{
			pCurve->SetX(j, pData->X(j));
			pCurve->SetY(j, 0.0);
		}
		if(pData->nData() < 2) continue;
		pCurve->SetFileName(pData->FileName());
		CString s1;
		s1.Format("Curve#%2d - %lg", i, m_Position[i]);
		pCurve->SetComment(s1 + pData->FileName());
		Lf = m_FWHMLorentz[i];
		Gf = m_FWHMGauss[i];
		if(Gf >= Lf) qw = Gf * m_QRange;
		else qw = Lf * m_QRange;
		Q1 = m_Position[i];
		Qmin = Q1 - qw;
		Qmax = Q1 + qw;
		i0 = 0;
		i1 = datan;
		for(j = 0 ; j < datan ; j++)
		{
			if(i0 == 0 && pCurve->X(j) >= Qmin) i0 = j;
			if(i1 == datan && pCurve->X(j) >= Qmax) i1 = j;
		}
		if(i0 < 0) i0 = 0;
		if(i0 > datan) continue;
		if(i1 > datan) i1 = datan;
		if(i1 < 0) continue;
		rf = Gf / Lf;
		coeL = rf * rf * 4.0;
		for(j = i0 ; j < i1 ; j++)
		{
			Q = pCurve->X(j);
			Qa = Q-Q1;
			xx1 = Qa / m_FWHMGauss[i];
			f1 = m_CTotal[i] * ((m_GaussFraction[i] * exp(-coeG * xx1*xx1)
						+(1.0-m_GaussFraction[i]) / (1.0 + coeL * xx1*xx1)));
			pCurve->SetY(j, f1);
			m_Total.SetY(j, m_Total.Y(j)+f1);
		}
		if(i == m_iCurrentPeak)
			pCurve->m_LineLineColor = RGB(0,0,0xff);
		else
			pCurve->m_LineLineColor = RGB(0,0xff,0);
	}

	m_BG.CalcMinMax();
	m_Total.CalcMinMax();
	for(i = 0 ; i < m_Curves.nGraphData() ; i++)
	{
		CGraphData *pGraph = m_Curves.GetGraphData(i);
		if(pGraph == NULL) continue;
		pGraph->CalcMinMax();
	}
	return 1;
}

int CPeakSearch::CalcPeakFitPeak(int iCurve)
{
	double f1, f2;
	double Qmin, Qmax, qw, Q, xx1, xx2, Q1, Q2, d;
	int i0, i1, i, j, k;
	double Qa;
	double val, CotQ1, CotQ2, coeL;
	const double coeG = 2.772588722;
	const double DegToRad = 0.01745329251944;
	const double RadToDeg = 57.295779514720003;
	double Lf, Gf, rf, fj[7];
	const int nAsym = 7;
	const double gAsym[] = {1.0, 4.0, 2.0, 4.0, 2.0, 4.0, 1.0};
	const double coefAsym = 1.0 / 18.0;

	CGraphData *pData = m_pOriginalData;
	if(pData == NULL) return -1;
	int datan = pData->nData();
	m_BG.SetDataNumber(datan);
	m_Total.SetDataNumber(datan);
	for(i = 0 ; i < datan ; i++)
	{
		m_BG.SetX(i, pData->X(i));
		m_BG.SetY(i, 0.0);
		m_Total.SetX(i, pData->X(i));
//		m_Total.SetY(i, 0.0);
	}
	m_Total.m_LineLineColor = RGB(0xff, 0, 0);
	m_Total.SetFileName(pData->FileName());
	m_Total.SetComment("(Total)" + pData->FileName());
	m_BG.m_LineLineColor = RGB(0, 0xff, 0xff);
	m_BG.SetFileName(pData->FileName());
	m_BG.SetComment("(Background)" + pData->FileName());
	if(m_QRange < 0.1) m_QRange = 0.1;
	for(i = 0 ; i < datan ; i++)
	{
		Q = m_BG.X(i);
		val = m_BGb[0] + (m_BGb[1] + (m_BGb[2] + (m_BGb[3] + m_BGb[4]*Q) * Q) * Q) * Q;
		m_BG.SetY(i, val);
		m_Total.SetY(i, val);
	}

	m_Curves.RemoveAllData();
	for(i = 0 ; i < nAsym ; i++)
	{
		fj[i] = (double)i / (double)(nAsym-1);
		fj[i] = fj[i] * fj[i];
	}
	for(i = 0 ; i < m_nPeaks ; i++)
	{
		CGraphData *pCurve = m_Curves.AddGraph(datan);
		if(pCurve == NULL)
		{
			CString s;
			s.Format("Can't create curve #%d with datan=%d", i, datan);
			MessageBox(NULL, s, "Error", MB_OK);
			return -1;
		}
		for(j = 0 ; j < datan ; j++)
		{
			pCurve->SetX(j, pData->X(j));
			pCurve->SetY(j, 0.0);
		}
		if(pData->nData() < 2) continue;
		pCurve->SetFileName(pData->FileName());
		CString s1;
		s1.Format("Curve#%2d - %lg", i, m_Position[i]);
		pCurve->SetComment(s1 + pData->FileName());
		Lf = m_FWHMLorentz[i];
		Gf = m_FWHMGauss[i];
		if(Gf >= Lf) qw = Gf * m_QRange;
		else qw = Lf * m_QRange;
		Q1 = m_Position[i];
		if(Q1 <= 0.0 || Q1 >= 180.0) continue;
		d = sin(Q1*0.5*DegToRad) / m_LKa1 * 2.0;
		Q2 = 2.0*RadToDeg * asin(m_LKa2*0.5 * d);
		Qmin = m_Position[i] - qw;
		Qmax = Q2 + qw;
		i0 = 0;
		i1 = 0;
		for(j = 0 ; j < datan ; j++)
		{
			if(i0 == 0 && pCurve->X(j) >= Qmin) i0 = j;
			if(i1 == 0 && pCurve->X(j) >= Qmax) i1 = j-1;
		}
/*
		i0 = (int)( (Qmin - pCurve->X(0)) / (pCurve->X(1)-pCurve->X(0)) );
		i1 = (int)( (Qmax - pCurve->X(0)) / (pCurve->X(1)-pCurve->X(0)) );
*/
		if(i0 < 0) i0 = 0;
		if(i0 > datan) continue;
		if(i1 > datan) i1 = datan;
		if(i1 <= 0) i1 = datan;//continue;
		rf = Gf / Lf;
		coeL = rf * rf * 4.0;
//		for(j = i0 ; j < i1 ; j++)
//			pCurve->SetY(j, 0);
		CotQ1 = 1.0 / tan(Q1 * DegToRad);
		CotQ2 = 1.0 / tan(Q2 * DegToRad);
		for(j = i0 ; j < i1 ; j++)
		{
			Q = pCurve->X(j);
			f1 = 0.0;
			for(k = 0 ; k < nAsym ; k++)
			{
				Qa = Q-Q1 + fj[k]*m_Aniso[i] * CotQ1;
				xx1 = Qa / m_FWHMGauss[i];
				f2 = (m_GaussFraction[i] * exp(-coeG * xx1*xx1)
						+(1.0-m_GaussFraction[i]) / (1.0 + coeL * xx1*xx1));
				Qa = Q-Q2 + fj[k]*m_Aniso[i] * CotQ2;
				xx2 = Qa / m_FWHMGauss[i];
				f2 = f2 + m_IRatio * (m_GaussFraction[i] * exp(-coeG * xx2*xx2)
						+(1.0-m_GaussFraction[i]) / (1.0 + coeL * xx2*xx2));
				f1 = f1 + f2 * gAsym[k];
			}
			f1 = f1 * coefAsym * m_CTotal[i];
			pCurve->SetY(j, f1);
			m_Total.SetY(j, m_Total.Y(j)+f1);
		}
		if(i == m_iCurrentPeak)
			pCurve->m_LineLineColor = RGB(0,0,0xff);
		else
			pCurve->m_LineLineColor = RGB(0,0xff,0);
	}

	m_BG.CalcMinMax();
	m_Total.CalcMinMax();
	for(i = 0 ; i < m_Curves.nGraphData() ; i++)
	{
		CGraphData *pGraph = m_Curves.GetGraphData(i);
		if(pGraph == NULL) continue;
		pGraph->CalcMinMax();
	}
	return 1;
}

int CPeakSearch::AddPeak(double position, double ctotal, double gaussfraction, 
							double fwhmglratio, double fwhm)
{
	if(m_nPeaks >= m_MaxPeaks-1) return m_nPeaks;
	int n = m_nPeaks;
	m_Position[m_nPeaks]      = position;
	m_CTotal[m_nPeaks]        = ctotal;
	m_GaussFraction[m_nPeaks] = gaussfraction;
	m_FWHMGLRatio[m_nPeaks]   = fwhmglratio;
	m_CGauss[m_nPeaks]        = ctotal * gaussfraction;
	m_FWHMGauss[m_nPeaks]     = fwhm * fwhmglratio;
	m_CLorentz[m_nPeaks]      = ctotal * (1-gaussfraction);
	m_FWHMLorentz[m_nPeaks]   = fwhm;
	m_Aniso[m_nPeaks]         = 0.0;

	if(m_SpectrumKind == PEAKSEARCH_SPECTRUMKIND_XRD ||
	   (m_SpectrumKind != PEAKSEARCH_SPECTRUMKIND_XRD && IsUsePeakFit(m_RamanLSQProgram)))
	{
		m_OptimizePosition[m_nPeaks]      = 1;
		m_OptimizeCTotal[m_nPeaks]        = 1;
		m_OptimizeGaussFraction[m_nPeaks] = 1;
		m_OptimizeFWHMGLRatio[m_nPeaks]   = 0;
		m_OptimizeCGauss[m_nPeaks]        = 0;
		m_OptimizeFWHMGauss[m_nPeaks]     = 1;
		m_OptimizeCLorentz[m_nPeaks]      = 0;
		m_OptimizeFWHMLorentz[m_nPeaks]   = 0;
		m_OptimizeAniso[m_nPeaks]         = 0;

		m_DiffPosition[m_nPeaks]      = 0.0;
		m_DiffCTotal[m_nPeaks]        = 0.0;
		m_DiffGaussFraction[m_nPeaks] = 0.0;
		m_DiffFWHMGLRatio[m_nPeaks]   = 0.0;
		m_DiffCGauss[m_nPeaks]        = 0.0;
		m_DiffFWHMGauss[m_nPeaks]     = 0.0;
		m_DiffCLorentz[m_nPeaks]      = 0.0;
		m_DiffFWHMLorentz[m_nPeaks]   = 0.0;
		m_DiffAniso[m_nPeaks]         = 0.0;
	}
	else
	{
		m_OptimizePosition[m_nPeaks]      = 1;
		m_OptimizeCTotal[m_nPeaks]        = 0;
		m_OptimizeGaussFraction[m_nPeaks] = 0;
		m_OptimizeFWHMGLRatio[m_nPeaks]   = 0;
		m_OptimizeCGauss[m_nPeaks]        = 1;
		m_OptimizeFWHMGauss[m_nPeaks]     = 1;
		m_OptimizeCLorentz[m_nPeaks]      = 1;
		m_OptimizeFWHMLorentz[m_nPeaks]   = 1;
		m_OptimizeAniso[m_nPeaks]         = 0;

		m_DiffPosition[m_nPeaks]      = 10.0;
		m_DiffCTotal[m_nPeaks]        = 10.0;
		m_DiffGaussFraction[m_nPeaks] = 0.3;
		m_DiffFWHMGLRatio[m_nPeaks]   = 0.3;
		m_DiffCGauss[m_nPeaks]        = 10.0;
		m_DiffFWHMGauss[m_nPeaks]     = 3.0;
		m_DiffCLorentz[m_nPeaks]      = 10.0;
		m_DiffFWHMLorentz[m_nPeaks]   = 3.0;
		m_DiffAniso[m_nPeaks]         = 0.5;
	}
	m_nPeaks++;
	return m_nPeaks;
}

int CPeakSearch::DeletePeak(int i)
{
	if(i < 0 || i >= m_nPeaks) return m_nPeaks;
	for(int ii = i ; ii < m_nPeaks-1 ; ii++)
	{
		m_Position[ii]     = m_Position[ii+1];
		m_CTotal[ii]       = m_CTotal[ii+1];
		m_GaussFraction[ii] = m_GaussFraction[ii+1];
		m_CGauss[ii]       = m_CGauss[ii+1];
		m_FWHMGauss[ii]    = m_FWHMGauss[ii+1];
		m_CLorentz[ii]     = m_CLorentz[ii+1];
		m_FWHMLorentz[ii]  = m_FWHMLorentz[ii+1];
		m_FWHMGLRatio[ii]  = m_FWHMGLRatio[ii+1];
		m_Aniso[ii]        = m_Aniso[ii];
		m_OptimizePosition[ii]     = m_OptimizePosition[ii+1];
		m_OptimizeCTotal[ii]       = m_OptimizeCTotal[ii+1];
		m_OptimizeGaussFraction[ii] = m_OptimizeGaussFraction[ii+1];
		m_OptimizeCGauss[ii]       = m_OptimizeCGauss[ii+1];
		m_OptimizeFWHMGauss[ii]    = m_OptimizeFWHMGauss[ii+1];
		m_OptimizeCLorentz[ii]     = m_OptimizeCLorentz[ii+1];
		m_OptimizeFWHMLorentz[ii]  = m_OptimizeFWHMLorentz[ii+1];
		m_OptimizeFWHMGLRatio[ii]  = m_OptimizeFWHMGLRatio[ii+1];
		m_OptimizeAniso[ii]        = m_OptimizeAniso[ii];
	}
	m_nPeaks--;
	return m_nPeaks;
}

int CPeakSearch::ReadPFCFile(LPCSTR fname, bool ReadSavedData)
{
	if(m_SpectrumKind == PEAKSEARCH_SPECTRUMKIND_XRD)
	{
		return ReadPeakFitPFCFile(fname, true, ReadSavedData);
	}
	else
	{
		if(IsPeakFitPFCFile(fname))
		{
			//MessageBox(NULL, "PeakFit", "aaa", MB_OK);
			return ReadPeakFitPFCFile(fname, false, ReadSavedData);
		}
		else
		{
			//MessageBox(NULL, "RmnFit", "aaa", MB_OK);
			return ReadRmnFitPFCFile(fname, ReadSavedData);
		}
	}
	return -1;
}

int CPeakSearch::ReadPeakFitPFCFile(LPCSTR fname, bool IsXRD, bool ReadSavedData)
{
	FILE *in;
	char str[256];
	int i;

	if((in = fopen(fname, "r")) == NULL) return -1;
	RemoveAll();
	fgets(str, 256, in);
	m_Title = str;
	m_Title.TrimLeft();
	m_Title.TrimRight();
	fgets(str, 256, in);
	m_DataPath = str;
	m_DataPath.TrimRight();
	m_DataPath.TrimLeft();
	char *cp = strrchr(str, '.');
	fgets(str, 256, in);
	if(IsXRD)
		m_XRDLSQProgram = str;
	else
		m_RamanLSQProgram = str;
	m_XRDLSQProgram.TrimRight();
	m_XRDLSQProgram.TrimLeft();
	m_RamanLSQProgram.TrimRight();
	m_RamanLSQProgram.TrimLeft();
	fgets(str, 256, in);
	m_LSQMethod = str;
	m_LSQMethod.TrimRight();
	m_LSQMethod.TrimLeft();
	fscanf(in, "%d %lg %lg", &m_MaxIter, &m_DumpingFactor, &m_ConvergenceCriteria);
	fscanf(in, "%lg %lg %lg %lg", &m_LKa1, &m_LKa2, &m_QRange, &m_IRatio);
	fscanf(in, "%lg %lg", &m_OutputRegionX0, &m_OutputRegionX1);
	fscanf(in, "%lg %lg", &m_OptRegionX0, &m_OptRegionX1);
	for(i = 0 ; i < 5 ; i++)
	{
		fscanf(in, "%lg %d %lg", &m_BGb[i], &m_OptimizeBGb[i], &m_DiffBGb[i]);
	}
	fscanf(in, "%d", &m_nPeaks);
	for(i = 0 ; i < m_nPeaks ; i++)
	{
		double height, gfrac;
		fscanf(in, "%lg %d %lg", &m_Position[i], &m_OptimizePosition[i], &m_DiffPosition[i]);
		fscanf(in, "%lg %d %lg", &height, &m_OptimizeCTotal[i], &m_DiffCTotal[i]);
		fscanf(in, "%lg %d %lg", &gfrac, &m_OptimizeGaussFraction[i], &m_DiffGaussFraction[i]);
		m_CTotal[i] = height;
		m_GaussFraction[i] = gfrac;
		m_CGauss[i] = height*gfrac;
		m_CLorentz[i] = height*(1-gfrac);
		m_OptimizeCGauss[i] = 0;
		m_OptimizeCLorentz[i] = 0;
		m_OptimizeGaussFraction[i] = m_OptimizeGaussFraction[i];
		fscanf(in, "%lg %d %lg", &m_FWHMGauss[i], &m_OptimizeFWHMGauss[i], &m_DiffFWHMGauss[i]);
		fscanf(in, "%lg %d %lg", &m_FWHMLorentz[i], &m_OptimizeFWHMLorentz[i], &m_DiffFWHMLorentz[i]);
		fscanf(in, "%lg %d %lg", &m_FWHMGLRatio[i], &m_OptimizeFWHMGLRatio[i], &m_DiffFWHMGLRatio[i]);
		fscanf(in, "%lg %d %lg", &m_Aniso[i], &m_OptimizeAniso[i], &m_DiffAniso[i]);
	}
	int ndata;
	if(ReadSavedData)
	{
		fscanf(in, "%d", &ndata);
		m_pOriginalData->SetDataNumber(ndata);
		for(i = 0 ; i < ndata ; i++)
		{
			double x, y;
			fscanf(in, "%lg %lg", &x, &y);
			m_pOriginalData->SetX(i, x);
			m_pOriginalData->SetY(i, y);
		}
	}
	m_pOriginalData->CalcMinMax();
	fclose(in);
	return 1;
}

int CPeakSearch::ReadRmnFitPFCFile(LPCSTR fname, bool ReadSavedData)
{
	FILE *in;
	char str[256];
	int i;

	if((in = fopen(fname, "r")) == NULL) return -1;
	RemoveAll();
	fgets(str, 256, in);
	m_Title = str;
	m_Title.TrimRight();
	m_Title.TrimLeft();
	m_DataPath = m_Title;
	fgets(str, 256, in);
	sscanf(str, "%d %d", &m_iPrt, &m_iPrtIntvl);
	fgets(str, 256, in);
	double y;
	sscanf(str, "%d %lg %lg", &m_MaxIter, &m_ConvergenceCriteria, &y);
	fgets(str, 256, in);
	sscanf(str, "%lg %lg", &m_OptRegionX0, &m_OptRegionX1);
	m_OutputRegionX0 = m_OptRegionX0;
	m_OutputRegionX1 = m_OptRegionX1;
	fgets(str, 256, in);
	sscanf(str, "%d", &m_BGOrder);
//	MessageBox(NULL, str, "BGOrder", MB_OK);
	for(i = 0 ; i <= m_BGOrder ; i++)
	{
		fgets(str, 256, in);
		sscanf(str, "%lg %d %lg", &m_BGb[i], &m_OptimizeBGb[i], &m_DiffBGb[i]);
//	MessageBox(NULL, str, "BG", MB_OK);
	}
	for( ; i < 5 ; i++)
	{
		m_BGb[i] = 0.0;
		m_OptimizeBGb[i] = 0;
	}
	if(m_BGOrder > 4) m_BGOrder = 4;
	fgets(str, 256, in);
	sscanf(str, "%d", &m_nPeaks);
//	MessageBox(NULL, str, "nPeaks", MB_OK);
	for(i = 0 ; i < m_nPeaks ; i++)
	{
		fgets(str, 256, in);
		sscanf(str, "%lg %d %lg", &m_Position[i], &m_OptimizePosition[i], &m_DiffPosition[i]);
//		MessageBox(NULL, str, "Pos", MB_OK);
		fgets(str, 256, in);
		sscanf(str, "%lg %d %lg", &m_CGauss[i], &m_OptimizeCGauss[i], &m_DiffCGauss[i]);
//		MessageBox(NULL, str, "CG", MB_OK);
		fgets(str, 256, in);
		sscanf(str, "%lg %d %lg", &m_FWHMGauss[i], &m_OptimizeFWHMGauss[i], &m_DiffFWHMGauss[i]);
//		MessageBox(NULL, str, "fwhmg", MB_OK);
		fgets(str, 256, in);
		sscanf(str, "%lg %d %lg", &m_CLorentz[i], &m_OptimizeCLorentz[i], &m_DiffCLorentz[i]);
//		MessageBox(NULL, str, "CL", MB_OK);
		fgets(str, 256, in);
		sscanf(str, "%lg %d %lg", &m_FWHMLorentz[i], &m_OptimizeFWHMLorentz[i], &m_DiffFWHMLorentz[i]);
//		MessageBox(NULL, str, "fwhml", MB_OK);
		m_CTotal[i] = m_CGauss[i] + m_CLorentz[i];
		m_OptimizeCTotal[i] = 0;
		m_GaussFraction[i] = m_CGauss[i] / m_CTotal[i];
		m_OptimizeGaussFraction[i] = 0;
		m_FWHMGLRatio[i] = m_FWHMGauss[i]/m_FWHMLorentz[i];
		m_OptimizeFWHMGLRatio[i] = 0;
		m_Aniso[i] = 0.0;
		m_OptimizeAniso[i] = 0;
	}
//	fgets(str, 256, in);
	fgets(str, 256, in);
	sscanf(str, "%lg %lg", &m_Sigma, &m_PercentSigma);
//	MessageBox(NULL, str, "Sigma", MB_OK);
	int ndata;
	if(ReadSavedData)
	{
		fscanf(in, "%d", &ndata);
		m_pOriginalData->SetDataNumber(ndata);
		for(i = 0 ; i < ndata ; i++)
		{
			double x, y;
			fscanf(in, "%lg %lg", &x, &y);
			m_pOriginalData->SetX(i, x);
			m_pOriginalData->SetY(i, y);
		}
		m_pOriginalData->CalcMinMax();
	}
	fclose(in);
	return 1;
}

int CPeakSearch::SavePFCFile(LPCSTR fname)
{
	if(m_SpectrumKind == PEAKSEARCH_SPECTRUMKIND_XRD)
		return SavePeakFitPFCFile(fname, m_XRDLSQProgram);
	else if(m_SpectrumKind == PEAKSEARCH_SPECTRUMKIND_RAMAN)
	{
		if(IsUsePeakFit(m_RamanLSQProgram))
			return SavePeakFitPFCFile(fname, m_RamanLSQProgram);
		else
			return SaveRmnFitPFCFile(fname);
	}
	return PEAKSEARCH_SPECTRUMKIND_INVALID;
}

int CPeakSearch::SaveRmnFitPFCFile(LPCSTR fname)
{
	FILE *out;
	int i;

	if(m_pOriginalData == NULL) return -1;
	if((out = fopen(fname, "w")) == NULL) 
	{
		CString s;
		s.Format("Can't open %s", fname);
		MessageBox(NULL, s, "Error in CPeakSearch::SaveRamanPFCFile", MB_OK);
		return -1;
	}

	fprintf(out, "%s\n", (LPCSTR)m_DataPath);
	fprintf(out, "%d %d\n", m_iPrt, m_iPrtIntvl);
	fprintf(out, "%d %lg %lg\n", m_MaxIter, m_ConvergenceCriteria, m_ConvergenceCriteria);
	fprintf(out, "%lg %lg\n", m_OptRegionX0, m_OptRegionX1);
	fprintf(out, "%d\n", m_BGOrder);
	for(i = 0 ; i <= m_BGOrder ; i++)
	{
		fprintf(out, "%lg %d %lg\n", m_BGb[i], m_OptimizeBGb[i], m_DiffBGb[i]);
	}
	fprintf(out, "%d\n", m_nPeaks);
	for(i = 0 ; i < m_nPeaks ; i++)
	{
		fprintf(out, "%lg %d %lg\n", m_Position[i], m_OptimizePosition[i], m_DiffPosition[i]);
		fprintf(out, "%lg %d %lg\n", m_CGauss[i], m_OptimizeCGauss[i], m_DiffCGauss[i]);
		fprintf(out, "%lg %d %lg\n", m_FWHMGauss[i], m_OptimizeFWHMGauss[i], m_DiffFWHMGauss[i]);
		fprintf(out, "%lg %d %lg\n", m_CLorentz[i], m_OptimizeCLorentz[i], m_DiffCLorentz[i]);
		fprintf(out, "%lg %d %lg\n", m_FWHMLorentz[i], m_OptimizeFWHMLorentz[i], m_DiffFWHMLorentz[i]);
	}
	fprintf(out, "%lg %lg\n", m_Sigma, m_PercentSigma);
	fprintf(out, "%d\n", m_pOriginalData->nData());
	for(i = 0 ; i < m_pOriginalData->nData() ; i++)
	{
		fprintf(out, "%lg %lg\n", m_pOriginalData->X(i), m_pOriginalData->Y(i));
	}
	fclose(out);
	return 1;
}

int CPeakSearch::SavePeakFitPFCFile(LPCSTR fname, LPCSTR LSQProgram)
{
	FILE *out;
	int i;
	CGraphData *pData = m_pOriginalData;
	if(pData == NULL) return -1;

	if((out = fopen(fname, "w")) == NULL) 
	{
		CString s;
		s.Format("Can't open %s", fname);
		MessageBox(NULL, s, "Error in CPeakSearch::SavePeakFitPFCFile", MB_OK);
		return -1;
	}
	fprintf(out, "%s\n", "title");
	fprintf(out, "%s\n", (LPCSTR)pData->FileName());
	fprintf(out, "%s\n", LSQProgram);
	fprintf(out, "%s\n", (LPCSTR)m_LSQMethod);
	fprintf(out, "%3d %6.4lf %10.4lg\n", m_MaxIter, m_DumpingFactor, m_ConvergenceCriteria);
	fprintf(out, "%10.6lf %10.6lf %10.4lf %10.4lf\n", m_LKa1, m_LKa2, m_QRange, m_IRatio);
	fprintf(out, "%10.6lf %10.6lf\n", m_OutputRegionX0, m_OutputRegionX1);
	fprintf(out, "%10.6lf %10.6lf\n", m_OptRegionX0, m_OptRegionX1);
	for(i = 0 ; i < 5 ; i++)
	{
		fprintf(out, "  %12.6lg %2d %12.6lg\n", m_BGb[i], m_OptimizeBGb[i], m_DiffBGb[i]);
	}
	fprintf(out, "%d\n", m_nPeaks);
	for(i = 0 ; i < m_nPeaks ; i++)
	{
		fprintf(out, "  %10.6lf %d %lg\n", m_Position[i], m_OptimizePosition[i], m_DiffPosition[i]);
		fprintf(out, "  %10.0lf %d %lg\n", m_CTotal[i], m_OptimizeCTotal[i], m_DiffCTotal[i]);
		fprintf(out, "  %10.4lf %d %lg\n", m_GaussFraction[i], m_OptimizeGaussFraction[i], m_DiffGaussFraction[i]);
		fprintf(out, "  %10.6lf %d %lg\n", m_FWHMGauss[i], m_OptimizeFWHMGauss[i], m_DiffFWHMGauss[i]);
		fprintf(out, "  %10.6lf %d %lg\n", m_FWHMLorentz[i], m_OptimizeFWHMLorentz[i], m_DiffFWHMLorentz[i]);
		fprintf(out, "  %10.6lf %d %lg\n", m_FWHMGLRatio[i], m_OptimizeFWHMGLRatio[i], m_DiffFWHMGLRatio[i]);
		fprintf(out, "  %10.6lf %d %lg\n", m_Aniso[i], m_OptimizeAniso[i], m_DiffAniso[i]);
	}

	fprintf(out, "%d\n", pData->nData());
	for(i = 0 ; i < pData->nData() ; i++)
	{
		fprintf(out, "  %10.4lf %14.4lf\n", pData->X(i), pData->Y(i));
	}
	fclose(out);
	return 1;
}
bool CPeakSearch::IsUsePeakFit(LPCSTR LSQProgram)
{
	LPCSTR cp = strrchr(LSQProgram, '\\');
	if(cp == NULL) cp = LSQProgram;
	char s[256];
	strcpy(s, cp);
	CharUpper(s);
	if(strstr(s, "PEAKFIT") != NULL) return true;
	return false;
}

bool CPeakSearch::IsPeakFitPFCFile(LPCSTR fname)
{
	FILE *in;
	if((in = fopen(fname, "r")) == NULL) return false;
	char str[256];
	fgets(str, 256, in);
	fgets(str, 256, in);
	fgets(str, 256, in);
	*str = 0;
	fgets(str, 256, in);
	fclose(in);
	if(strnicmp(str, "marq", 4) == 0) return true;
	return false;
}

int CPeakSearch::SetYAxisType(int type)
{
	m_BG.SetYAxisType(type);
	m_Total.SetYAxisType(type);
	for(int i = 0 ; i < m_Curves.nGraphData() ; i++)
	{
		CGraphData *pdata = m_Curves.GetGraphData(i);
		if(pdata == NULL) continue;
		pdata->SetYAxisType(type);
	}
	return 1;
}
