import sys
import numpy as np
from numpy import exp
import pandas as pd
from scipy.optimize import minimize


#==========================================
# Parameters
#==========================================
#nelder-mead    Downhill simplex
method = "nelder-mead"

maxiter = 100
tol = 1.0e-5

#==========================================
# Source parameters to be fitted
#==========================================
infile = 'peak.xlsx'
nx = None

# optimization parameters
x0 = [1.3,  0.6, 0.1]


argv = sys.argv
n = len(argv)
if n >= 2:
    method = argv[1]
if n >= 3:
    infile = argv[2]
if n >= 4:
    x0[0] = float(argv[3])
if n >= 5:
    x0[1] = float(argv[4])
if n >= 6:
    x0[2] = float(argv[5])


#==========================================
# functions
#==========================================
def ycal(x, params):
    I0 = params[0]
    x0 = params[1]
    w  = params[2]
    if w < 0.001:
        w = 0.001

    a = 0.832554611 / w
    X = a * (x - x0)

    return I0 * exp(-X * X)

def ycal_list(xd, params):
    y = []
    for i in range(len(xd)):
        y.append(ycal(xd[i], params))
    return y

def CalS2(params):
    global xd, yd, nx
    S2 = 0.0
    for i in range(nx):
        yc = ycal(xd[i], params)
        d  = yd[i] - yc
        S2 += d * d
    return S2


def main():
    global nx
    global xd, yd, xiter, yfmin

    print("")
    print("Peak fitting by python scipy.optimize")
    print(f"method {method}")
    print(f"input file {infile}")
    print(f"initial values {x0}")

    print("")
    print(f"Read [{infile}]")
    df = pd.read_excel(infile, engine = 'openpyxl')
    labels = df.columns.to_list()
    xd = df[labels[0]].to_numpy()
    yd = df[labels[1]].to_numpy()
    nx = len(xd)
    xrange = [min(xd), max(xd)]

    yini = ycal_list(xd, x0)

    res = minimize(CalS2, x0, method = method, tol = tol, options = {'maxiter':maxiter, "disp":True})
    print("")
    print("Optimization finished")
    print(res)
    print("")
    print(f"Converence status: {res.message}")
    print(f"optimized parameters: {res.x}")
    print(f"Residual S^2: {res.fun}")


if __name__ == "__main__":
    main()
